/* File: ViewMatrix.java
 * Contains: ACP3 ViewMatrix
 * Author: Austin Tate <a.tate@ed.ac.uk>
 * Created: Fri Jan 16 11:13:15 1998
 * Updated: Fri Aug 11 16:44:44 2000 by Jeff Dalton
 * Copyright: (c) 1998, AIAI, University of Edinburgh
 */

/* ACP3 ViewMatrix Panel.
 * Columns used for COA-1..N and use indexed 1..N. Index 0 is used for Overall
 * Rows 1..N used for process steps.  Row 0 not used
 *
 * Main methods for use externally:
 * 
 * vmClear        clear the ViewMatrix display and reinitialise data structures
 *                Currently also loads the default process TIE 97-1 model
 *
 * loadMatrix     loads in the process model, levels information and
 *                desired appearance
 *
 * paintMatrix    to allow for a test interface to ACP3. Can be used to
 *                manually set up demo states or load demo state files.
 *
 * setColumnHeading overrides the heading used for any column (until reset)
 *
 * setDrawOn      sets screen updating of the matrix on and off.  This allows
 *                changes using all the following routines to be hidden (false)
 *                until the drawing is set back on (true)
 *
 * setNumberOfCOA is used to indicate how many COAs to display.
 *                Use first to make columns and image available.
 *
 * getLabelIndex  is used to get the row number that goes with a given label.
 * 
 * setStatus      is used to set status (ViewColor.STATUS_??) and text of a
 *                specific process step overall (col=0) or in a nominated COA
 *            
 * setIssue       is used to set status (ViewColor.STATUS_??) and text of the
 *                "issues remaining" row overall (col=0) or in a nominated COA
 *            
 * setExpanded    to mark a suitable process step as showing expanded
 *
 * setContracted  to mark a suitable process step as showing contracted
 */

package ix.acp3;

import java.awt.*;
import java.awt.image.*;
import java.util.*;
import ix.util.*;

class ViewMatrix extends Panel implements Runnable {

    private Thread vmThread;

    private boolean vmIsInitialised; //false until vmInit called
    private Image triDown, triRight; //for triangle expansion icons
    private Image offsc;       //offscreen drawing area
    private Graphics osg;    
    private Container parent;  //access to parent for container validation
    private ImageCanvas vmc;   //ViewMatrix canvas
    private VMQueue vmq;       //ViewMatrix event queue
    private int osWidth, osHeight; //Current off screen image width and height
    //appearance control
    private boolean issuesOnShow, overallOnShow; //set in loadMatrix
    private String overallLabel, coaLabelRoot;   //set in loadMatrix

    //configuration values set in advance - but alterable if followed by reset
    int topMargin=0, bottomMargin=0,
                leftMargin=0, rightMargin=0,
                cellAdornment=3, //pixels from boundary to cell text clip area
                triWidth=12, triHeight=12, triToLabelGap=2,                
                indentStep=6, limitOnLabelWidth=300,
                gapToOverall=20, gapToIssues=8;

    private int leftMouseEdge=-1, rightMouseEdge=-1, //tri-icon event detection
                topMouseEdge=-1, bottomMouseEdge=-1,
                //mouse zone is set initially in case handeEvent is called
                lastRowHighlight=0; // last row in currennt display that was
                                    // was highlighted in mouseMoved, 0 if none
    private int maxLabelWidth, //set dynamically in setLabelInfo   
                maxRows,       //set dynamically in constructor
                coaOnShow,     //COAs currently on show, excluding overall (0)
                rowsOnShow;    //row number currently on show, excluding header
    // flag to indicate whether immediate screen update is required
    // for external method calls
    private boolean drawingIsOn=true;
    // values dependent on Default Font - set in vmInit
    private static int overallWidth, coaWidth, rowHeight;

    private String[] processLabel; //set in loadMatrix
                                   //0th element for issues remaining element

    // processExpandStatus: 0 for not expandable, 0 means expandable
    //                      1 for expandable not expanded, 2 for expanded
    //note must be 0 or 1 initially, 2 is assigned when the class is in use
    //and is reset back to 0 or 1 in setProcessInfo
    private int[] processExpandStatus;
    // processLevel gives indent level - 0=top level, 1=level 1, etc.
    private int[] processLevel;  // 0th element used for issues

    /* private */ Hashtable processLabelIndexTable; // set in loadMatrix
            
    //cellStatus initialised in setProcessInfo and altered in createNewMatrix
    private int[][] cellStatus;
    //columnLabel and columnColor initialised in setProcessInfo
    //and altered in createNewMatrix
    private String[] columnLabel;
    private Color[] columnColor;

    //cellLabel initialised in setProcessInfo and altered in createNewMatrix
    //index 0 is label used on Issues row
    private String[][] cellLabel;
    //processLabelWidth set up by creator
    private int[] processLabelWidth;
    //set up by setProcessInfo and then dynamically as expand tri-icons are used
    private int[] processAtRow; //mapping of row numbers to process step on row
    
    int widthSpaceNormal;
    FontMetrics fmNormal, fmBold;
                
    ViewMatrix (Container parent) {
        super();
        this.parent=parent;
        //get tri-icons
        triDown=WaitImage.waitGetImage("IMG/tri-down.gif",this);
        triRight=WaitImage.waitGetImage("IMG/tri-right.gif",this);
        reset(); //for safety
        //next line should really be done outside of the class
        defaultProcessModel(); //does a 2nd reset via loadMatrix
    }

    private void reset() {
        Debug.noteln("ViewMatrix reset performed");
	if (vmq == null) vmq=new VMQueue(); else vmq.clear();
        maxRows=0; //initialise to no rows (no process model)
        coaOnShow=0; //initialise to no COAs        
        rowsOnShow=0;   //initialise number of rows on show        
        maxLabelWidth=0; //initialise max label
        // empty initial structures to allow for no rows and no COAs
        cellStatus=new int[1][1]; //i.e. no COAs and no Rows
        cellLabel=new String[1][1];
        processExpandStatus=new int[1];
        processLevel=new int[1];
        processLabel=new String[1];
        processLabel[0]="";
        processLabelWidth=new int[1];
        columnLabel=new String[1]; //just label column 0 as default
        columnLabel[0]="";
        columnColor=new Color[1];
        columnColor[0]=Color.white;
        //initialise mouse zone to no area
        leftMouseEdge=-1;
        rightMouseEdge=-1;
        topMouseEdge=-1;
        bottomMouseEdge=-1;        
        lastRowHighlight=0;  // no highlights yet

        vmIsInitialised=false;
        //note must now call vmInit() at some convenient place
        //to complete initialisation - but after graphics context is known
        //done in vmInit called from createNewMatrix on first use
    }

    public void vmClear () {
        // Handle the request in vmThread.
	// Queue the event.
        vmq.send(new VMEvent(VMEvent.VM_EVENT,1));
	resetFinished.nextMessage();
    }

    public void loadMatrix (String[] pLabel, int[] pLevel,
                            boolean showIssues, boolean showOverall,
                            String overallL, String coaLRoot) {
        int lastLevel;
        int len=pLabel.length-1;
        Debug.noteln("loadMatrix called.");
        if ((((len+1)!=pLevel.length) && (len>0)) || (pLevel[0]!=0)) {
            Debug.noteln("loadMatrix parameters incorrect - must have "+
                         "length > 0, arranys be same length, "+
                         "and pLevel[0]=0");
        } else {
            reset(); //forces clean up of data structures
            maxRows=len;
            processLabel= new String[maxRows+1];
            processLabel[0]="";
            processLevel=new int[maxRows+1];
            processLevel[0]=0; //index 0 used for issues
            processLabel[0]=pLabel[0];
	    processLabelIndexTable=new Hashtable();
            processExpandStatus=new int[maxRows+1];
            processExpandStatus[0]=0; //index 0 used for issues
            lastLevel=0;
            for (int i=1; i<=maxRows; i++) {
                processLabel[i]=pLabel[i];
                processLevel[i]=pLevel[i];
                processExpandStatus[i]=0; //not expandable
		processLabelIndexTable.put(pLabel[i], new Integer(i));
                if (pLevel[i]>lastLevel) {
                    if (i>1) {
                        processExpandStatus[i-1]=1; //expandable
                    }             
                // limit level down change to a step of 1
                   if (pLevel[i]!=lastLevel+1) {
                        Debug.noteln("loadMatrix levels error at "+i);
                        processLevel[i]=lastLevel+1; //limit down change to 1
                    }
                }

                lastLevel=processLevel[i];
	    }
            // now create data structures that are dependent on process model
            processAtRow=new int[maxRows+1];
            processLabelWidth=new int[maxRows+1];
            // next two set in createNewCOA
            cellStatus=new int[maxRows+1][1];
            cellLabel=new String[maxRows+1][1];
            // set appearance
            issuesOnShow=showIssues;
            overallOnShow=showOverall;
            overallLabel=overallL;
            coaLabelRoot=coaLRoot;
        }
    }

    public void defaultProcessModel () {
        String[] pLabel=new String[]
	    {"Issues Remaining", //0th element used for issues row label
            "develop JFACC guidance",                             // 1
                "formalize JFACC objectives",                     // 2
                "identify COGs",                                  // 3   
                "decompose objectives to tasks",                  // 4
                "assign MOMs",                                    // 5
            "develop targets",                                    // 6
            "formalize planning advice",                          // 7
            "develop COA",                                        // 8
                "create COA",                                     // 9
                    "execute plan-cycle",                         //10
                        "create COA to one more level",           //11
                        "approve COA",                            //12
                        "reject COA",                             //13
                    "approve final COA",                          //14
                "assign weapons systems",                         //15
                "create support plan",                            //16
                    "execute support-plan-cycle",                 //17
                        "create support plan to one more level",  //18
                        "compose schedule",                       //19
                        "approve support plan",                   //20
                        "reject support plan",                    //21
                    "approve final support plan",                 //22
                "review COA",                                     //23
                    "critique COA by simulation",                 //24
                    "critique COA by resource-comparison",        //25
                    "critique COA by bed-down evaluation",        //26
                    "critique COA by quality measures",           //27
                "monitor COA-creation process",                   //28
            "decide to explore alternative COA(s)",               //29
            "decide to brief on COA(s)",                          //30
            "brief on COA(s)",                                    //31
                "create briefing",                                //32
                "communicate briefing"                            //33
            };
        int[] pLevel={0,  //0th element used for issues row - must be 0
            0,1,1,1,1,0,0,0,1,2,3,3,3,2,1,1,2,3,3,3,3,2,1,2,2,2,2,1,0,0,0,1,1
            };
        loadMatrix(pLabel,pLevel,true,true,"   Overall   ","COA-");
    }

    public void setColumnHeading (int c, String s, Color col) {
        columnLabel[c]=s;
        columnColor[c]=col;
    }

    public void setDrawOn (boolean flag) {
        if (drawingIsOn!=flag) {
            drawingIsOn=flag;
            if (flag) { //if not on before and set on - redraw matrix
               redrawMatrix();
               vmc.repaint();
            }
	}
    }

    // Runnable requirements follow

    // executed after init for each thread, before anything else
    public void start () {
      if (vmThread==null) {
         vmThread=new Thread(this,"ViewMatrix Thread");
         vmThread.start();
      };
      Thread.yield();
    }

//     public void stop () {
//       if (vmThread!=null) {
//          vmThread.stop();
//          vmThread=null; // set to null to indicate vmThread is not running
//       };
//     }

    // executed after start for a thread
    public void run () {
	Object message;
        VMEvent vme;
        if (Thread.currentThread()==vmThread) {
            while (true) {
                // next line waits until there is something in the Queue
		message = vmq.nextMessage();
		// Debug.noteln("Message:", message);
		if (message instanceof ProcessStatusUpdate) {
		    handleStatusUpdate((ProcessStatusUpdate)message);
		}
		else {
		    vme=(VMEvent)message;
		    switch (vme.ev) {
		    case  VMEvent.VM_EVENT: //code 1 only used just now
			handleReset();
			break;
		    case VMEvent.TRI_EVENT:
			handleTriEvent(vme.code);
			break;
		    case VMEvent.PM_EVENT:
			handlePaintMatrix(vme.code);
			break;
		    default:
			Debug.noteln("Incorrect VMEvent="+vme.ev);
			break;
		    }
	        }
	    }
        }
    }

    // end of Runnable required methods

    public MessageQueue resetFinished = new MessageQueue();

    public void handleReset() {
	if (osg!=null) {
	    clearContainer(this);
	    parent.validate();
	}
	//reset(); //ensure in if line below removed
	//next line should really be done outside
	//of the class
	defaultProcessModel(); //does a reset
	resetFinished.send(new Boolean(true));
    }

    public void statusUpdate(ProcessStatusUpdate up) {
	vmq.send(up);
    }

    private void handleStatusUpdate(ProcessStatusUpdate up) {

	setDrawOn(false);

	if (up.containsCompleteState) {
	    // Setting the number of COA to 0 will have the side-effect
	    // of resetting all the status values to BLANK.  This isn't
	    // really necessary (modulo bugs), except for an empty model,
	    // because the update contains a complete description.
	    // setNumberOfCOA doesn't set status in the overall column.  /\/
	    setNumberOfCOA(0);
	    // If the model's empty, make sure the overall column is 
	    // initialized.
	    if (up.statusChanges.isEmpty()) {
		// First set everything in the overall column to BLANK.
		for (int i = 1; i <= maxRows; i++)
		    setStatus(i,0,ViewColor.STATUS_BLANK,null);
		// The top entry in the overall column is POSSIBLE.
		// The process model can't easily tell us this, because
		// it doesn't know enough until there's a COA.  /\/
		setStatus(1, 0, ViewColor.STATUS_POSSIBLE, null);
		// Now there's one issue in the overall column.
		setIssue(0,ViewColor.STATUS_NA,"1");
	    }
	    // Start with nothing expanded.
	    resetToContracted();
	    // Expand some things
	    setExpanded(getLabelIndex("develop COA"));
	    setExpanded(getLabelIndex("create COA"));
	    setExpanded(getLabelIndex("create support plan"));
	}

	setNumberOfCOA(up.numberOfCoas);

	// Node status changes.
	for (Enumeration ce = up.statusChanges.elements();
	     ce.hasMoreElements();) {
	    StatusChange change = (StatusChange)ce.nextElement();
	    int i = getLabelIndex(change.action);
	    String text = change.statusText();
	    if (change.coa >= 1) {
		setStatus(i, change.coa, change.status, text);
	    }
	    else if (change.coa == StatusChange.COA_OVERALL)
		setStatus(i, 0, change.status, text);
	    else {
		Debug.assert(change.coa == StatusChange.COA_NONE);
		// Put the status value in every entry on the row.
		for (int coa = 1; coa <= up.numberOfCoas; coa++) {
		    setStatus(i, coa, change.status, text);
		}
	    }
	}

	// Issue counts
	Debug.assert(up.issueCounts.length == up.numberOfCoas + 1);
	for (int coa = 0; coa < up.issueCounts.length; coa++)
	    setIssue(coa,
		     ViewColor.STATUS_NA,
		     String.valueOf(up.issueCounts[coa]));

	setDrawOn(true);

    }

    private void handlePaintMatrix(int code) {
      switch (code) {
      case 0:
	setDrawOn(false);
	setNumberOfCOA(0);
	resetToContracted(); //for speed - replaces setContracteds
	// setExpanded(1); simpler to have closed at first
	// Overall
	setStatus(1,0,ViewColor.STATUS_POSSIBLE,null);
	setStatus(2,0,ViewColor.STATUS_POSSIBLE,null);
	for (int i=3; i<=maxRows; i++) {
	  setStatus(i,0,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,0,ViewColor.STATUS_NA,"N/A");
	setIssue(0,ViewColor.STATUS_NA,"1");
	setDrawOn(true);
	break;
      case 1:
	setDrawOn(false);
	setNumberOfCOA(1);
	resetToContracted(); //for speed - replaces setContracteds
	// COA-1
	for (int i=1; i<=28; i++) {
	  setStatus(i,1,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(29,1,ViewColor.STATUS_NA,"N/A");
	setStatus(30,1,ViewColor.STATUS_POSSIBLE,null);
	for (int i=31; i<=maxRows; i++) {
	  setStatus(i,1,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,1,ViewColor.STATUS_NA,"N/A");
	setStatus(29,1,ViewColor.STATUS_NA,"N/A");
	setIssue(1,ViewColor.STATUS_NA,"1");
	// Overall
	for (int i=1; i<=28; i++) {
	  setStatus(i,0,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(29,0,ViewColor.STATUS_POSSIBLE,null);
	setStatus(30,0,ViewColor.STATUS_POSSIBLE,"COA-1");
	for (int i=31; i<=maxRows; i++) {
	  setStatus(i,0,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,0,ViewColor.STATUS_NA,"N/A");
	setIssue(0,ViewColor.STATUS_NA,"2");
	setDrawOn(true);
	break;
      case 2:
	setDrawOn(false);
	setNumberOfCOA(2);
	resetToContracted(); //for speed - replaces setContracteds
	setExpanded(1);
	// COA-1
	for (int i=1; i<=28; i++) {
	  setStatus(i,1,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(29,1,ViewColor.STATUS_NA,"N/A");
	setStatus(30,1,ViewColor.STATUS_POSSIBLE,null);
	for (int i=31; i<=maxRows; i++) {
	  setStatus(i,1,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,1,ViewColor.STATUS_NA,"N/A");
	setStatus(29,1,ViewColor.STATUS_NA,"N/A");
	setIssue(1,ViewColor.STATUS_NA,"1");
	// COA-2
	setStatus(1,2,ViewColor.STATUS_POSSIBLE,null);
	setStatus(2,2,ViewColor.STATUS_POSSIBLE,null);
	for (int i=3; i<=maxRows; i++) {
	  setStatus(i,2,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,2,ViewColor.STATUS_NA,"N/A");
	setStatus(29,2,ViewColor.STATUS_NA,"N/A");
	setIssue(2,ViewColor.STATUS_NA,"1");
	// Overall
	setStatus(1,0,ViewColor.STATUS_POSSIBLE,null);
	setStatus(2,0,ViewColor.STATUS_POSSIBLE,null);
	for (int i=3; i<=28; i++) {
	  setStatus(i,0,ViewColor.STATUS_BLANK,null);
	};
	setStatus(29,0,ViewColor.STATUS_COMPLETE,null);
	setStatus(30,0,ViewColor.STATUS_POSSIBLE,"COA-1");
	for (int i=31; i<=maxRows; i++) {
	  setStatus(i,0,ViewColor.STATUS_BLANK,null);
	};
	setIssue(0,ViewColor.STATUS_NA,"2");
	setDrawOn(true);
	break;
      case 3:
	setDrawOn(false);
	setNumberOfCOA(3);
	resetToContracted(); //for speed - replaces setContracteds
	setExpanded(1);
	// COA-1
	for (int i=1; i<=28; i++) {
	  setStatus(i,1,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(29,1,ViewColor.STATUS_NA,"N/A");
	setStatus(30,1,ViewColor.STATUS_POSSIBLE,null);
	for (int i=31; i<=maxRows; i++) {
	  setStatus(i,1,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,1,ViewColor.STATUS_NA,"N/A");
	setStatus(29,1,ViewColor.STATUS_NA,"N/A");
	setIssue(1,ViewColor.STATUS_NA,"1");
	// COA-2
	setStatus(1,2,ViewColor.STATUS_POSSIBLE,null);
	setStatus(2,2,ViewColor.STATUS_POSSIBLE,null);
	for (int i=3; i<=maxRows; i++) {
	  setStatus(i,2,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,2,ViewColor.STATUS_NA,"N/A");
	setStatus(29,2,ViewColor.STATUS_NA,"N/A");
	setIssue(2,ViewColor.STATUS_NA,"1");
	// COA-3
	setStatus(1,3,ViewColor.STATUS_POSSIBLE,null);
	setStatus(2,3,ViewColor.STATUS_POSSIBLE,null);
	for (int i=3; i<=maxRows; i++) {
	  setStatus(i,3,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,3,ViewColor.STATUS_NA,"N/A");
	setStatus(29,3,ViewColor.STATUS_NA,"N/A");
	setIssue(3,ViewColor.STATUS_NA,"1");
	// Overall
	setStatus(1,0,ViewColor.STATUS_POSSIBLE,null);
	setStatus(2,0,ViewColor.STATUS_POSSIBLE,null);
	for (int i=3; i<=28; i++) {
	  setStatus(i,0,ViewColor.STATUS_BLANK,null);
	};
	setStatus(29,0,ViewColor.STATUS_COMPLETE,null);
	setStatus(30,0,ViewColor.STATUS_POSSIBLE,"COA-1");
	for (int i=31; i<=maxRows; i++) {
	  setStatus(i,0,ViewColor.STATUS_BLANK,null);
	};
	setIssue(0,ViewColor.STATUS_NA,"3");
	setDrawOn(true);
	break;
      case 4:
	setDrawOn(false);
	setNumberOfCOA(3);
	resetToContracted(); //for speed - replaces setContracteds
	setExpanded(8);
	// setExpanded(9);
	setExpanded(23);
	// COA-1
	for (int i=1; i<=28; i++) {
	  setStatus(i,1,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(29,1,ViewColor.STATUS_NA,"N/A");
	setStatus(30,1,ViewColor.STATUS_POSSIBLE,null);
	for (int i=31; i<=maxRows; i++) {
	  setStatus(i,1,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,1,ViewColor.STATUS_NA,"N/A");
	setStatus(29,1,ViewColor.STATUS_NA,"N/A");
	setIssue(1,ViewColor.STATUS_NA,"1");
	// COA-2
	for (int i=1; i<=7; i++) {
	  setStatus(i,2,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(8,2,ViewColor.STATUS_EXECUTING,null);
	for (int i=9; i<=22; i++) {
	  setStatus(i,2,ViewColor.STATUS_COMPLETE,null);
	};
	for (int i=23; i<=27; i++) {
	  setStatus(i,2,ViewColor.STATUS_EXECUTING,null);
	};
	setStatus(28,2,ViewColor.STATUS_EXECUTING,null);
	for (int i=29; i<=maxRows; i++) {
	  setStatus(i,2,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,2,ViewColor.STATUS_NA,"N/A");
	setStatus(29,2,ViewColor.STATUS_NA,"N/A");
	setStatus(24,2,ViewColor.STATUS_COMPLETE,null);
	setIssue(2,ViewColor.STATUS_NA,"2");
	// COA-3
	for (int i=1; i<=7; i++) {
	  setStatus(i,3,ViewColor.STATUS_COMPLETE,null);
	};
	for (int i=8; i<=11; i++) {
	  setStatus(i,3,ViewColor.STATUS_EXECUTING,null);
	};
	for (int i=12; i<=24; i++) {
	  setStatus(i,3,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,3,ViewColor.STATUS_NA,"N/A");
	for (int i=26; i<=27; i++) {
	  setStatus(i,3,ViewColor.STATUS_BLANK,null);
	};
	setStatus(28,3,ViewColor.STATUS_EXECUTING,null);
	for (int i=29; i<=maxRows; i++) {
	  setStatus(i,3,ViewColor.STATUS_BLANK,null);
	};
	setStatus(29,3,ViewColor.STATUS_NA,"N/A");
	setIssue(3,ViewColor.STATUS_NA,"2");
	// Overall
	for (int i=1; i<=7; i++) {
	  setStatus(i,0,ViewColor.STATUS_COMPLETE,null);
	};
	for (int i=8; i<=11; i++) {
	  setStatus(i,0,ViewColor.STATUS_EXECUTING,null);
	};
	for (int i=12; i<=22; i++) {
	  setStatus(i,0,ViewColor.STATUS_BLANK,null);
	};
	setStatus(23,0,ViewColor.STATUS_EXECUTING,null);
	setStatus(24,0,ViewColor.STATUS_BLANK,null);
	setStatus(25,0,ViewColor.STATUS_EXECUTING,"COA-1,2");
	for (int i=26; i<=28; i++) {
	  setStatus(i,0,ViewColor.STATUS_EXECUTING,null);
	};
/*      for (int i=26; i<=27; i++) {
	  setStatus(i,0,ViewColor.STATUS_BLANK,null);
	};
	setStatus(28,0,ViewColor.STATUS_EXECUTING,null);
*/
	setStatus(29,0,ViewColor.STATUS_BLANK,null);
	setStatus(30,0,ViewColor.STATUS_POSSIBLE,"COA-1");
	for (int i=31; i<=maxRows; i++) {
	  setStatus(i,0,ViewColor.STATUS_BLANK,null);
	};
	setIssue(0,ViewColor.STATUS_NA,"6");
	setDrawOn(true);
	break;
      case 5:
	setDrawOn(false);
	setNumberOfCOA(3);
	resetToContracted(); //for speed - replaces setContracteds
	setExpanded(8);
	setExpanded(23);
	// COA-1
	for (int i=1; i<=28; i++) {
	  setStatus(i,1,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(29,1,ViewColor.STATUS_NA,"N/A");
	setStatus(30,1,ViewColor.STATUS_POSSIBLE,null);
	for (int i=31; i<=maxRows; i++) {
	  setStatus(i,1,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,1,ViewColor.STATUS_NA,"N/A");
	setStatus(29,1,ViewColor.STATUS_NA,"N/A");
	setIssue(1,ViewColor.STATUS_NA,"1");
	// COA-2
	for (int i=1; i<=7; i++) {
	  setStatus(i,2,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(8,2,ViewColor.STATUS_EXECUTING,null);
	for (int i=9; i<=22; i++) {
	  setStatus(i,2,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(23,2,ViewColor.STATUS_EXECUTING,null);
	setStatus(24,2,ViewColor.STATUS_COMPLETE,null);
	for (int i=25; i<=27; i++) {
	  setStatus(i,2,ViewColor.STATUS_EXECUTING,null);
	};
	setStatus(28,2,ViewColor.STATUS_COMPLETE,null);
	for (int i=29; i<=maxRows; i++) {
	  setStatus(i,2,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,2,ViewColor.STATUS_NA,"N/A");
	setStatus(29,2,ViewColor.STATUS_NA,"N/A");
	setStatus(27,2,ViewColor.STATUS_COMPLETE,null);
	setIssue(2,ViewColor.STATUS_NA,"1");
	// COA-3
	for (int i=1; i<=7; i++) {
	  setStatus(i,3,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(8,3,ViewColor.STATUS_EXECUTING,null);
	for (int i=9; i<=22; i++) {
	  setStatus(i,3,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(23,3,ViewColor.STATUS_EXECUTING,null);
	setStatus(24,3,ViewColor.STATUS_COMPLETE,null);
	for (int i=25; i<=27; i++) {
	  setStatus(i,3,ViewColor.STATUS_EXECUTING,null);
	};
	setStatus(28,3,ViewColor.STATUS_COMPLETE,null);
	for (int i=29; i<=maxRows; i++) {
	  setStatus(i,3,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,3,ViewColor.STATUS_NA,"N/A");
	setStatus(29,3,ViewColor.STATUS_NA,"N/A");
	setIssue(3,ViewColor.STATUS_NA,"2");
	// Overall
	for (int i=1; i<=7; i++) {
	  setStatus(i,0,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(8,0,ViewColor.STATUS_EXECUTING,null);
	for (int i=9; i<=22; i++) {
	  setStatus(i,0,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(23,0,ViewColor.STATUS_EXECUTING,null);
	setStatus(24,0,ViewColor.STATUS_COMPLETE,null);
	setStatus(25,0,ViewColor.STATUS_EXECUTING,"COA-1,2,3");
	for (int i=26; i<=27; i++) {
	  setStatus(i,0,ViewColor.STATUS_EXECUTING,null);
	};
	setStatus(28,0,ViewColor.STATUS_COMPLETE,null);
	setStatus(29,0,ViewColor.STATUS_BLANK,null);
	setStatus(30,0,ViewColor.STATUS_POSSIBLE,"COA-1");
	for (int i=31; i<=maxRows; i++) {
	  setStatus(i,0,ViewColor.STATUS_BLANK,null);
	};
	setIssue(0,ViewColor.STATUS_NA,"5");
	setDrawOn(true);
	break;
      case 6:
	setDrawOn(false);
	setNumberOfCOA(3);
	resetToContracted(); //for speed - replaces setContracteds	
	setExpanded(8);
	setExpanded(23);
	// COA-1
	for (int i=1; i<=28; i++) {
	  setStatus(i,1,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(29,1,ViewColor.STATUS_NA,"N/A");
	setStatus(30,1,ViewColor.STATUS_POSSIBLE,null);
	for (int i=31; i<=maxRows; i++) {
	  setStatus(i,1,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,1,ViewColor.STATUS_NA,"N/A");
	setStatus(29,1,ViewColor.STATUS_NA,"N/A");
	setIssue(1,ViewColor.STATUS_NA,"1");
	// COA-2
	for (int i=1; i<=28; i++) {
	  setStatus(i,2,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(29,2,ViewColor.STATUS_NA,"N/A");
	setStatus(30,2,ViewColor.STATUS_POSSIBLE,null);
	for (int i=31; i<=maxRows; i++) {
	  setStatus(i,2,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,2,ViewColor.STATUS_NA,"N/A");
	setStatus(29,2,ViewColor.STATUS_NA,"N/A");
	setIssue(2,ViewColor.STATUS_NA,"1");
	// COA-3
	for (int i=1; i<=28; i++) {
	  setStatus(i,3,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(29,3,ViewColor.STATUS_NA,"N/A");
	setStatus(30,3,ViewColor.STATUS_POSSIBLE,null);
	for (int i=31; i<=maxRows; i++) {
	  setStatus(i,3,ViewColor.STATUS_BLANK,null);
	};
	setStatus(25,3,ViewColor.STATUS_NA,"N/A");
	setStatus(29,3,ViewColor.STATUS_NA,"N/A");
	setIssue(3,ViewColor.STATUS_NA,"1");
	// Overall
	for (int i=1; i<=24; i++) {
	  setStatus(i,0,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(25,0,ViewColor.STATUS_COMPLETE,"COA-1,2,3");
	for (int i=26; i<=28; i++) {
	  setStatus(i,0,ViewColor.STATUS_COMPLETE,null);
	};
	setStatus(29,0,ViewColor.STATUS_POSSIBLE,null);
	setStatus(30,0,ViewColor.STATUS_POSSIBLE,"COA-1,2,3");
	for (int i=31; i<=maxRows; i++) {
	  setStatus(i,0,ViewColor.STATUS_BLANK,null);
	};
	setIssue(0,ViewColor.STATUS_NA,"4");
	setDrawOn(true);
	break;
      }
    }

    static final class NonclearingImageCanvas extends ImageCanvas {
	// This class added by jd.
	NonclearingImageCanvas(Image img) {
	    super(img);
	}
	public void update(Graphics g) {
	    // Don't clear -- avoid the flash
	    paint(g);
	}
    }

    //initialise ViewMatrix variables that need peer to be present
    private void vmInit () {
        //offsc gives an initial graphics context - changed in each call
        //to createNewMatrix
        offsc=WaitImage.waitCreateImage(12,12,this);
        osg=offsc.getGraphics();
        //create canvas - only done once
        //used to be: vmc=new ImageCanvas(offsc);
	vmc=new NonclearingImageCanvas(offsc);
        vmc.setCanvasImage(offsc);
        //single canvas added to panel
        add(vmc);

        //set up font information
        osg.setFont(DefaultFont.bold);
        fmBold=osg.getFontMetrics();
        osg.setFont(DefaultFont.normal);
        fmNormal=osg.getFontMetrics();
        widthSpaceNormal=fmNormal.stringWidth(" ");
        //initialise process (row) labels and to set up for no COAs on display
	setProcessInfo();
        //setup reasonable widths and heights for columns
        //needs setProcessInfo information
        coaWidth=calcWidthOneLine(coaLabelRoot+"N", fmBold)
                 +(2*cellAdornment);
        overallWidth=calcWidthOneLine(overallLabel, fmBold)
                 +(2*cellAdornment);
        rowHeight=fmBold.getHeight()+4;
        if (rowHeight<triHeight) rowHeight=triHeight+4;
        vmIsInitialised=true;
    }

    private int calcWidthLabel (int i) {
        FontMetrics fm;
        if (processExpandStatus[i]>0) fm=fmBold;
        else fm=fmNormal;
        //indent
        return triWidth+triToLabelGap+
               +((processLevel[i]*indentStep)*widthSpaceNormal)
               +(calcWidthOneLine(processLabel[i], fm));
    }
    
    // routine only for use internally to be helpful to demos where it is
    // known that a new matrix is being set up

    private void resetToContracted () {
        for (int i=1; i<=maxRows; i++) { //ignore issues for this purpose
            if (processExpandStatus[i]==2) processExpandStatus[i]=1;
        }
        rowsOnShow=0;
        for (int i=1; i<=maxRows; i++) { //ignore issues for this purpose
            if (processLevel[i]==0) {
                rowsOnShow=rowsOnShow+1;
                processAtRow[rowsOnShow]=i;
            }                           
        }
    }

    private void setProcessInfo () {
        int temp;
        maxLabelWidth=0; //initialise max label width found
        columnLabel=new String[1]; //just label column 0 as default
        columnLabel[0]=overallLabel;
        columnColor=new Color[1];
        columnColor[0]=Color.gray;
        for (int i=0; i<=maxRows; i++) {
            //row 0 for issues, column 0 for overall
            cellStatus[i][0]=ViewColor.STATUS_BLANK;
            cellLabel[i][0]=null;
            temp=calcWidthLabel(i);
            //keep actual label width - for text clipping;
            processLabelWidth[i]=temp;
            //now set actual label cell width
            temp=temp+(2*cellAdornment);            
            if (temp>maxLabelWidth) maxLabelWidth=temp;
            // initialise to have all top level process steps (level=0) on show
        }
        resetToContracted();
        if (maxLabelWidth>limitOnLabelWidth) maxLabelWidth=limitOnLabelWidth;
    }
    
    private void setMouseZone() {      
        //set Tri_Icon event detection zone
        leftMouseEdge=leftMargin;
        rightMouseEdge=leftMargin+maxLabelWidth;
        topMouseEdge=topMargin+rowHeight;
        bottomMouseEdge=topMargin+(rowHeight*(rowsOnShow+1));
        // Debug.noteln("Mouse zone set to "+leftMouseEdge+","+topMouseEdge+","+
        //                      rightMouseEdge+","+bottomMouseEdge);
    }
 
    //ACP3 demonstration interface
    public void paintMatrix(int code) {
        //handle the request in vmThread.  Queue the event
        vmq.send(new VMEvent(VMEvent.PM_EVENT,code));
    }    

    private void redrawMatrix () {
        if (osg!=null && drawingIsOn) {
            //clear screen first for safety
            osg.setColor(Color.white);
            osg.fillRect(0,0,osWidth,osHeight);
            //COA column headings
            for (int c=1; c<=coaOnShow; c++) {
                drawCell(-1,c,columnColor[c],columnLabel[c],
                          DefaultFont.bold,Color.black);
            }        
            //overall column heading (index column 0)
            if (overallOnShow)
                drawCell(-1,0,columnColor[0],columnLabel[0],
                                  DefaultFont.bold,Color.black);
            //draw rest of matrix - all rows from row 1 downward
            drawRowsFrom(1);
        }
    }
         
    //createNewMatrix makes a new off screen image for a coaNumNew X RowsOnShow
    //Matrix and informs canvas containing it    

    private void createNewMatrix(int coaNumNew) {
        // if necessary set up font and vmc information, etc.
        if (coaNumNew<0) {
            Debug.noteln("createNewMatrix parameter must be >=0");
            return;
        }
        if (!vmIsInitialised) vmInit();
        Debug.noteln("createNewMatrix COAs="+coaOnShow+"->"+coaNumNew+
                     " Rows="+rowsOnShow);
        if (coaNumNew>coaOnShow) {
            int[][] newCellStatus=new int[maxRows+1][coaNumNew+1];
            String[][] newCellLabel=new String[maxRows+1][coaNumNew+1];
            String[] newColumnLabel=new String[coaNumNew+1];
            Color[] newColumnColor=new Color[coaNumNew+1];
            // copy previous status over to new arrays
            for (int j=0; j<=coaOnShow; j++) {
                for (int i=0; i<=maxRows; i++) {
                    newCellStatus[i][j]=cellStatus[i][j];
                    newCellLabel[i][j]=cellLabel[i][j];
                }                
                newColumnLabel[j]=columnLabel[j];
                newColumnColor[j]=columnColor[j];
            }
            cellStatus=newCellStatus;
            cellLabel=newCellLabel;
            columnLabel=newColumnLabel;
            columnColor=newColumnColor;
        } //else reuse smaller part of existing array
        //initialise extra columns now available (if relevant)
        //note coaOnShow may already be >= coaNumNew
        for (int j=coaOnShow+1; j<=coaNumNew; j++) {
            for (int i=0; i<=maxRows; i++) {
                cellStatus[i][j]=ViewColor.STATUS_BLANK; // =0
                cellLabel[i][j]=null;
                columnLabel[j]=coaLabelRoot+j;
                columnColor[j]=Color.gray;
            }            
        }
        coaOnShow=coaNumNew;  //set number of COAs to display
        osWidth=leftMargin+maxLabelWidth+(coaOnShow*coaWidth)+rightMargin;
        if (overallOnShow) {
            osWidth=osWidth+gapToOverall+overallWidth;
	}
        osHeight=topMargin+((rowsOnShow+1)*rowHeight)+bottomMargin;
        if (issuesOnShow) {
            osHeight=osHeight+gapToIssues+rowHeight; 
        }
        setMouseZone();
        offsc=WaitImage.waitCreateImage(osWidth,osHeight,this);
        osg=offsc.getGraphics();
        //change canvas on panel and ensure enclosing components are relaid out
        vmc.setCanvasImage(offsc);
        redrawMatrix();
        this.invalidate();
        parent.validate();
    }

    //handle an expand or contract event on row r

    private void handleTriEvent(int r) {
        int processNum=processAtRow[r];
        // Debug.noteln("handleTriEvent proc="+processNum+" row="+r);
        boolean expand;        
        if (processExpandStatus[processNum]==1) {
            expand=true;
            processExpandStatus[processNum]=2;
            // Debug.noteln("do expand");
        } else {
            if (processExpandStatus[processNum]==2) {
                expand=false; //i.e. contract required
                processExpandStatus[processNum]=1;
                // Debug.noteln("do contract");
            } else {
                // Debug.noteln("no action");
                return; //no action if not expandable
            }
        }
        //quick screen update of affected row for good user feedback
        drawRow(r);
        // could clear down - but update is fast enough to not require this
        // clearFromRow(r+1);
        vmc.repaint();            
        //now recompute what is on show and change data structures
        if (expand) {
            openUpFrom(r);
            //new larger offscreen image needed
            //so recreate and redraw whole thing
            createNewMatrix(coaOnShow); //mouse zone is set here anyway
            //repaint done automatically in createNewMatrix
        } else {
            //contract
            //reuse existing image as its large enough already
            closeUpFrom(r);
            //now redraw lines below r
            drawRowsFrom(r+1);
            setMouseZone(); //to new smaller area            
        }              
    }

    //clear offscreen image from row r downwards

    /* not used
    private void clearFromRow (int r) {            
        if (rowsOnShow>=r) {
            // Debug.noteln("clearFromRow "+r+" to "+rowsOnShow);
            int top=topMargin+(r*rowHeight);            
            osg.setColor(Color.white);
            osg.fillRect(0,top,osWidth,osHeight-top);
        }        
    }
    */
                
    //closeUpFrom will remove from display all sub-process steps below
    //the process on display at r

    private void closeUpFrom(int r) {
        if (r<rowsOnShow) {
            int level=processLevel[processAtRow[r]];
            //step until next rows's process is at or above same level
            //as processNum
            int row=r;
            while (row<rowsOnShow &&
                   processLevel[processAtRow[row+1]]>level) {
        	row++;
            };
            //row is either =r or last row to be closed up
            //close up from r+1 to row
            int closeUp=row-r;  //number of rows to close up
            if (closeUp>0) {
                // Debug.noteln("Closing up from "+(r+1)+" to "+row+"="+closeUp);
                for (int i=r+1; i<=(rowsOnShow-closeUp); i++) {
                    processAtRow[i]=processAtRow[i+closeUp];
                };
                rowsOnShow=rowsOnShow-closeUp;
            } else {
                // Debug.noteln("No closeup needed");
            }
	}
    }

    //openUpFrom will add to the display all sub-process steps below
    //the process on display at r

    private void openUpFrom(int r) {
        int processNum=processAtRow[r];
        int level=processLevel[processNum];
        int proc=processNum;        
        int addIn=0;
        int[] addInProc=new int[maxRows]; //maximum number that can be added
					  //is actually maxRows-1
                                          //do not use index 0
        int tempStatus, levelForProc;
        //step until next process is at or above same level as processNum
        //add next level sub-processes and those below that that are expanded
        while (proc<maxRows && processLevel[proc+1]>level) {
            proc++;
            addIn++;
            addInProc[addIn]=proc;
            tempStatus=processExpandStatus[proc];
            if (tempStatus==1) { //expandable but not expanded - skip over
                levelForProc=processLevel[proc];
                while (proc<maxRows && processLevel[proc+1]>levelForProc) {
                    proc++;
                }
            } //else if tempStatus=0 or 2 just carry on to next cycle
        };

        if (addIn>0) {
            if ((rowsOnShow+addIn)<=maxRows) {
                // Debug.noteln("Opening up to add "+addIn+" processes ");
                for (int i=rowsOnShow+addIn; i>(r+addIn); i--) {
                    processAtRow[i]=processAtRow[i-addIn];
                };
                for (int i=1; i<=addIn; i++) {
                    processAtRow[r+i]=addInProc[i];
                };
                rowsOnShow=rowsOnShow+addIn;
            } else{
               // Debug.noteln("Error in openUpFrom as maxRows exceeded.");
               rowsOnShow=maxRows; //reset to prevent bounds errors
            }
        }        
    }

    //draw each row from stated row (from) to end of rows on show
    //used with from=1 initially, and subsequently as tri-icons are used

    private void drawRow (int r) {
        int tempStatus;
        String tempString;        //used to hold String to be printed in cell
        int tempProcessNum;        
        //row process name and tri-icon labels
        if (drawingIsOn) {
            if ((r>0) || ((r==0) && issuesOnShow)) {
                drawLabel(r);
                if (r==0) tempProcessNum=0;
                else tempProcessNum=processAtRow[r];
                //now each cell entry (excluding overall=0)
                for (int c=0; c<=coaOnShow; c++) {
                    tempStatus=cellStatus[tempProcessNum][c];
                    tempString=cellLabel[tempProcessNum][c];
                    drawCell(r,c,ViewColor.statusColor[tempStatus],
                              tempString,DefaultFont.normal,
                              textColorForStatus(tempStatus));
                }
                //now handle overall
                if (overallOnShow) {
                    tempStatus=cellStatus[tempProcessNum][0];
                    tempString=cellLabel[tempProcessNum][0];
                    drawCell(r,0,ViewColor.statusColor[tempStatus],
                              tempString,DefaultFont.normal,
                              textColorForStatus(tempStatus));
		}
	    }
	}
    }

    private void drawRowsFrom (int from) {        
       //note from may be rowsOnShow+1 in some cases - just draw issues then
       if (drawingIsOn) {
            //clear offscreen image from row r downwards
            int top=topMargin+(from*rowHeight);            
            osg.setColor(Color.white);
            osg.fillRect(0,top,osWidth,osHeight-top);
            if (rowsOnShow>=from) {
                for (int r=from; r<=rowsOnShow; r++) drawRow(r);
	    }
            if (issuesOnShow) drawRow(0); // for issues
        }    
    }
                
    // processOnShowAtRow returns  >0 giving row where process is on show
    //                             -1 if not on show
    //                              0 is not used
                                    
    private int processOnShowAtRow (int processNum) {
        if ((processNum>=1) && (processNum<=maxRows)) {
            for (int i=1; i<=rowsOnShow; i++) {
                if (processAtRow[i]==processNum) {
                    return i;
                }
            }
	}
        return -1; //means not on show
    }
    
    //set status for a given COA, or for overall using coaNum=0

    // select text display color for given status

    private Color textColorForStatus (int status) {
        if (status==ViewColor.STATUS_NA) return Color.darkGray;
        else return Color.black; 
    }

    public int getLabelIndex(String label) {
	Integer index = (Integer)processLabelIndexTable.get(label);
	if (index == null)
	    throw new IllegalArgumentException("unknown label " + label);
	else
	    return index.intValue();
    }

    public void setStatus(int processNum, int coaNum, int status, String str) {
        if ((processNum>=0) && (processNum<=maxRows) &&
            (coaNum>=0) && (coaNum<=coaOnShow)) {
            cellStatus[processNum][coaNum]=status;
            cellLabel[processNum][coaNum]=str;
            if (drawingIsOn) {
                //see if this process step is on display
                int temp=processOnShowAtRow(processNum);        
                if (temp>0) {
                    Color ctemp=ViewColor.statusColor[status];
                    drawCell(temp, coaNum, ctemp, str,
                              DefaultFont.normal, textColorForStatus(status));
                }
	    }
	} else {
            Debug.noteln("setStatus proc="+processNum+", coa="+
                         coaNum+" out of range");
	}
    }    

    public void setIssue(int coaNum, int status, String str) {
        cellStatus[0][coaNum]=status;
        cellLabel[0][coaNum]=str;
        if (drawingIsOn) {
            Color ctemp=ViewColor.statusColor[status];
            drawCell(0, coaNum, ctemp, str, DefaultFont.normal,
                      textColorForStatus(status));
	}
    }    
    
    public void setNumberOfCOA(int coaNum) {
         createNewMatrix(coaNum);  //set up the matrix
    }

    //set a suitable process step as showing expanded
    public void setExpanded (int processNum) {
        if (processExpandStatus[processNum]==1) {
            int temp=processOnShowAtRow(processNum);
            if (temp>0) handleTriEvent(temp);            
            else processExpandStatus[processNum]=2;
        }        
    }
                
    //set a suitable process step as showing contracted
    public void setContracted (int processNum) {
        if (processExpandStatus[processNum]==2) {
            int temp=processOnShowAtRow(processNum);
            if (temp>0) handleTriEvent(temp);            
            else processExpandStatus[processNum]=1;
        }        
    }
                
    private void drawLabel(int r) {
        int processNum=processAtRow[r];
        int indent=indentStep*processLevel[processNum];
        int indentw=indent*widthSpaceNormal;
        String indents ="";
        for (int i=0; i<indent; i++) indents=indents+" ";
        int top, left, texth, textBaseLine;
        int triState, triIconOffset;
        Font fn;
        FontMetrics fm;
        // set drawing origin
        if (r==0) top=topMargin+(rowsOnShow*rowHeight)+gapToIssues+rowHeight;
        else top=topMargin+(r*rowHeight);        
        left=leftMargin;
        osg.setColor(Color.black);
        osg.drawRect(left+1,top+1,maxLabelWidth-2,rowHeight-2); 
        osg.setColor(Color.white);
        osg.fillRect(left+2,top+2,maxLabelWidth-3,rowHeight-3);
        osg.setClip(left+cellAdornment,top+cellAdornment,
                    (maxLabelWidth-1)-cellAdornment,
                    (rowHeight-1)-cellAdornment);
        // put up tri-icon if required
        triState=processExpandStatus[processNum];
        if (triHeight<rowHeight) triIconOffset=(rowHeight-triHeight)/2;
        else triIconOffset=2;               
        if (triState==1) {
            osg.drawImage(triRight,left+cellAdornment,top+triIconOffset,this);
        } else {
            if (triState==2) {
                osg.drawImage(triDown,left+cellAdornment,top+triIconOffset,this);
            }            
        }

        // draw text with idents spaces leading left aligned in the box
        osg.setColor(Color.black);
        if (indents!="") {
            osg.setFont(DefaultFont.normal);
            texth=fmNormal.getHeight();
            textBaseLine=texth+((rowHeight-texth)/2)-2;            
            osg.drawString(indents,
                           left+cellAdornment+triWidth+triToLabelGap,
                           top+textBaseLine);
        };
        if (processExpandStatus[processNum]>0) {
            fn=DefaultFont.bold;
            fm=fmBold;
        } else {
            fn=DefaultFont.normal;
            fm=fmNormal;
        };
        osg.setFont(fn);            
        texth=fm.getHeight();
        textBaseLine=texth+((rowHeight-texth)/2)-2;            
        osg.drawString(processLabel[processNum],
                       left+cellAdornment+triWidth+triToLabelGap+indentw,
                       top+textBaseLine);
      
        // reset clipping to whole area for safety
        osg.setClip(0,0,osWidth,osHeight);
    }
        
    //draw cell with cclr background and text in specified font and color
    //row = -1 for COA/Overall headings, 0 for issues
    //      N for normal process step

    //Note no checks on whether the cell should be visible with
    //drawingIsOn, issuesOnShow or overallOnShow

    private void drawCell(int rCode, int c, Color cclr,
                          String text, Font fn, Color tclr) {
        int textw, texth, textOffset, textBaseline;
        FontMetrics fm;
        int left, top, colWidth;                      
        // set drawing origin
        switch (rCode) {
            case -1: top=topMargin;
                     break;
            case  0: top=topMargin+((rowsOnShow+1)*rowHeight)+gapToIssues;
                     break;
            default: top=topMargin+(rCode*rowHeight);
                     break;
        }        
        if (c==0) {
            left=leftMargin+maxLabelWidth+
                        ((coaOnShow)*coaWidth)+gapToOverall;
            colWidth=overallWidth;
        } else {
            left=leftMargin+maxLabelWidth+((c-1)*coaWidth);
            colWidth=coaWidth;
        }
        osg.setColor(Color.black);
        osg.drawRect(left+1,top+1,colWidth-2,rowHeight-2); 
        osg.setColor(cclr);
        osg.fillRect(left+2,top+2,colWidth-3,rowHeight-3);

        // draw text centred in the box space
        if (text!=null) {
            if (!text.equals("")) {
                osg.setColor(tclr);
                osg.setFont(fn);            
                fm=osg.getFontMetrics();
                textw=calcWidthOneLine(text,fm);
                texth=fm.getHeight();
                textBaseline=texth+((rowHeight-texth)/2)-2;            
                if (textw<(colWidth-(2*cellAdornment))) {
                    textOffset=(int)((colWidth-4)-textw)/2;
                } else {
                    textOffset=0;
                }
                osg.setClip(left+cellAdornment,top+cellAdornment,
                            (colWidth-1)-cellAdornment,
                            (rowHeight-1)-cellAdornment);
                osg.drawString(text,left+cellAdornment+textOffset,
                                  top+textBaseline);
                // reset clipping to whole area
                osg.setClip(0,0,osWidth,osHeight);
            }
        }
        //could reset color to black for safety
        //osg.setColor(Color.black);                
    }

    // next 3 methods could usefully be moved to ix.util.Util
    
    private int calcWidthOneLine(String string, FontMetrics fm) {
        return fm.stringWidth(string);
    }

    // calcWidthMultiLine not yet used or tested
    /*
    private int calcWidthMultiLine(String string, FontMetrics fm) {
        int w = 0;
        String temp = null;
        //calc multi-line width
        StringTokenizer st1 = new StringTokenizer(string, "\n");
        for  (Enumeration  e  = st1; e.hasMoreElements()  ;)  {
            temp = (String) e.nextElement();
            if (fm.stringWidth(temp) > w)
            w = fm.stringWidth(temp);
        }
        return w;
    }
    */

    private void clearContainer (Container cont) {
        Component compList[];
        compList=cont.getComponents();
        for (int i=0; i<compList.length; i++) {
            cont.remove(compList[i]);
        };
        cont.invalidate(); //now need to revalidate enclosing container
    }

    //note that we detect mouse positions wrt the canvas - not to the off
    //screen image.  Need to relocate mouse position for this.

    public boolean mouseDown(Event ev, int x, int y) {
        int r;
        if (vmc!=null) {
            Point vmcLocation=vmc.getLocation();
            // Debug.noteln("Mouse down at "+x+","+y+" vmc at "+
            //               vmcLocation.x+","+vmcLocation.y);
            int vmcX=x-vmcLocation.x;
            int vmcY=y-vmcLocation.y;
            if (vmcX>=leftMouseEdge && vmcX<=rightMouseEdge
                      && vmcY>=topMouseEdge && vmcY<=bottomMouseEdge) {
                if (ev.metaDown()==false) {
                    //compute row
                    r=(int)((vmcY+1)/rowHeight); //+1 to ensure no divide by 0
                    if (r>0 && r<=rowsOnShow) {
                        // Debug.noteln("Tri-Icon event detected");
    	                // handleTriEvent(r); in vmThread
                        vmq.send(new VMEvent(VMEvent.TRI_EVENT,r));
                    }
                } else {
                    Debug.noteln("Right button not in use");
                }
	    }
        } else {
            Debug.noteln("Mouse events ignored until Image created");
        }        
        return true;
    }

    // Comment out mouseMove and mouseExit to turn off highlighting.
    // Matrix flashes due to full canvas repaints on vmc as mouse moved.
    /*
    public boolean mouseMove(Event ev, int x, int y) {
        int r;
        boolean altered=false;
        if (vmc!=null && drawingIsOn) {
            Point vmcLocation=vmc.getLocation();
            int vmcX=x-vmcLocation.x;
            int vmcY=y-vmcLocation.y;
            if (vmcX>=leftMouseEdge && vmcX<=rightMouseEdge
                     && vmcY>=topMouseEdge && vmcY<=bottomMouseEdge) {
                //compute row
                r=(int)((vmcY+1)/rowHeight); //+1 to ensure no divide by 0
                if (r>0 && r<=rowsOnShow) {
                    if (lastRowHighlight>0  && lastRowHighlight!=r &&
                          lastRowHighlight<=rowsOnShow) { //<- always true!
                        drawRow(lastRowHighlight); //repaint it normally
                    }
                    osg.setXORMode(Color.gray);
                    drawLabel(r);
                    lastRowHighlight=r;
                    osg.setPaintMode();
                    altered=true;
                }
	    }
           if (altered) vmc.repaint();
        }
        return true;
    }
    */

    /*
    public boolean mouseExit(Event ev, int x, int y) {
        boolean altered=true;
        if (vmc!=null && drawingIsOn) {
            if (lastRowHighlight>0) {
                drawRow(lastRowHighlight); //repaint it normally
                lastRowHighlight=0;
                vmc.repaint();
            }
        }
        return true;
    }
    */

}
  
