/* File: AutoTester.java
 * Contains: A class for simulating Kqml input
 * Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Created: April 1998
 * Updated: Thu Jun 22 15:34:15 2000 by Jeff Dalton
 * Copyright: (c) 1998, AIAI, University of Edinburgh
 */

package ix.ip2.coax;

import java.util.*;
import java.io.*;

import ix.util.*;
import ix.util.lisp.*;

/**
 * A class that can be used to replay recorded messages for testing
 * and demonstration.
 */

public class AutoTester extends Thread {

    protected static Vector allAutoTesters = new Vector();

    public static int defaultDelay = 5000; // milliseconds between messages
    public int maxDelay = 10000; 	   // milliseconds between messages
    public int minDelay = 0;		   // milliseconds between messages

    public static boolean askUserBeforeSend = false;

    public LList testMessages;

    public AutoTester(LList testMessages) {
	this.testMessages = testMessages;
 	allAutoTesters.addElement(this);
    }

    public static void init() {

	defaultDelay = Parameters.getInt("autotester-delay", defaultDelay);

	askUserBeforeSend = Parameters.getBoolean
	    ("autotester-ask-user-before-send");
    }

    public void run() {

	// Here's where we send the messages that constitute a test.
	Debug.noteln("AutoTester will send", testMessages);

	long time = 0;

	for (LList ms = testMessages; ms != Lisp.NIL; ms = ms.cdr()) {

	    String event = (String)ms.car();

	    Debug.noteln("\n---------- AutoTester sending", event);

	    if (askUserBeforeSend) {
		Util.askLine("--- press return to send");
	    }
	    else {
		long delay = defaultDelay;
		if (delay > maxDelay) {
		    Debug.noteln("Delay would be " + delay / 1000.0
				 + " seconds.");
		    Debug.noteln("Using " + maxDelay / 1000.0
				 + " seconds instead.");
		    delay = maxDelay;
		}
		if (delay > 0) {
		    try { Thread.sleep(delay); }
		    catch (InterruptedException e) {}
		}
	    }

	    // Here's where we actually "send" it.
	    CoAX_IP2.interpreter.interpret(event);

	}

	// The test is complete from our POV.
	Debug.noteln("\n\n- - - - - - - - - - - - - - - - - - - -\n");
	Debug.noteln("Autotester sent", testMessages);
	Debug.noteln("");

    }

    /*
     * Reset
     */

    public static void stopAll() {
	for (Enumeration e = allAutoTesters.elements();
	     e.hasMoreElements();) {
	    AutoTester t = (AutoTester)e.nextElement();
	    if (t.isAlive())
		t.stop();
	}
    }

    /*
     * Test table, etc.
     */

    protected static Vector testNames = new Vector();

    protected static Hashtable testTable = new Hashtable() {
	public Object put(Object key, Object value) {
	    testNames.addElement(key); // to remember the order in which added
	    return super.put(key, value);
	}
    };

    static {
	testTable.put("Default messages", "");
	testTable.put("Generate messages", "");
	testTable.put("Generate message permutation", "");
	// testTable.put("Ask for a file name","");
    }

    public static Enumeration testNames() {
	return testNames.elements();
    }

    public static void runTest(final String testName) {

	// New thread because we're called from a MainFrame menu-slection
	// and hence in an AWT thread.
	new Thread() {
	    public void run() {
		Debug.noteln("Test requested", testName);
		String filename = (String)testTable.get(testName);
		Debug.assert(filename != null);

		LList messages = Lisp.NIL;
		if (testName == "Default messages") {
		    messages =
			Seq.toList(Seq.elements(defaultMessageSequence));
		}
		else if (testName == "Generate messages") {
		    messages = generateTestMessageList();
		}
		else if (testName == "Generate message permutation") {
		    permuteSuccessors = true;
		    messages = generateTestMessageList();
		}
		else {
		    Debug.warn("Unknown test: " + testName);
		}

		// Send some messages.
		new AutoTester(messages).start();

	    }

	}.start();

    }

    /* 
     * Order of messages:
     *
     *  Start --> S1M1a --> S1M2a --> S1M3a --> S1M4a
     *              |         |         |         |
     *              V         V         V         V
     *            S1M1b --> S1M2b --> S1M3b     S1M4b
     *                                  |         ^
     *                                  V         |
     *                                  I1a ----> I1b
     *
     * John Levine, 9th June 2000.
     */

    static String[][] messageOrderSpecs = new String[][] {
        {"Start", "S1M1a"},
	{"S1M1a", "S1M1b", "S1M2a"},
	{"S1M1b", "S1M2b"},
        {"S1M2a", "S1M2b", "S1M3a"},
	{"S1M2b", "S1M3b"},
	{"S1M3a", "S1M3b", "S1M4a"},
        {"S1M3b", "I1a"},
        {"S1M4a", "S1M4b"},
        {"I1a", "I1b"},
        {"I1b", "S1M4b"},
	// Not sure the rest is right.
        {"S1M4b", "S2M1a"},
        {"S2M1a", "S2M1b", "S2M2a"},
        {"S2M1b", "S2M2b"},
        {"S2M2a", "S2M2b"},
        {"S2M2b"}};

    static String[] defaultMessageSequence = new String[] {
	"S1M1a", "S1M1b", "S1M2a", "S1M2b", "S1M3a", "S1M3b",
	"I1a", "I1b",
	"S1M4a", "S1M4b", "S2M1a", "S2M1b", "S2M2a", "S2M2b"
    };

    static Hashtable messageSuccessors = new Hashtable();

    public static LList defaultMessageList =
        Seq.toList(Seq.elements(defaultMessageSequence));

    // /\/: Having this global variable is a terrible way to do it.
    static boolean permuteSuccessors = false;

    static {
	for (int i = 0; i < messageOrderSpecs.length; i++) {
	    LList spec = Seq.toList(Seq.elements(messageOrderSpecs[i]));
	    String from = (String)spec.car();
	    LList to = spec.cdr();
	    messageSuccessors.put(from, to);
	}
    }

    static LList generateTestMessageList() {
	ListCollector result = new ListCollector();
	walkSampleTestMessages("Start", result.elementPusher()); 
	LList messages = result.contents();
	Debug.assert(((String)messages.car()).equals("Start"));
	return messages.cdr();	// minus "Start"
    }

    static void walkSampleTestMessages(String at, Function1 f) {
	// Think "topological sort".
	Hashtable marks = new Hashtable();
	messageWalk(at, f, marks);
    }

    static void messageWalk(String at, Function1 f, Hashtable marks) {
	if (marks.get(at) == null) {
	    marks.put(at, "start");
	    LList next = (LList)messageSuccessors.get(at);
	    // Can put them in some other order at this point.
	    if (permuteSuccessors)
		next = next.permute();
	    for (; next != Lisp.NIL; next = next.cdr()) {
		messageWalk((String)next.car(), f, marks);
	    }
	    f.funcall(at);
	    // Could change mark to "finish" for debugging
	}
    }

}
